/*
 * Copyright 2025 WebAssembly Community Group participants
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef interpreter_interpreter_h
#define interpreter_interpreter_h

#include "store.h"
#include "support/result.h"

namespace wasm {

class Interpreter {
public:
  // TODO: Methods to instantiate modules.
  // TODO: Methods to run exported functions.

  Result<> addInstance(std::shared_ptr<Module> wasm);
  std::vector<Literal> runTest(Expression* root);
  std::vector<Literal> run();

private:
  interpreter::WasmStore store;
  friend class InterpreterImpl;

  Result<> instantiate(interpreter::Instance& instance);
};

} // namespace wasm

#endif // interpreter_interpreter_h
